(** * Additional structure for call-by-value models *)

Require Import model_structure.

Set Implicit Arguments.
Unset Strict Implicit.

Module sforces_cbv (ks : Kripke_structure).
  Export ks.
  Module ks_monad := (Kripke_structure_monad ks).
  Export ks_monad.

  (** [sforces] is the strong forcing relation for call-by-value models *)
  Fixpoint sforces (w:K)(annot:bool)(A:formula) {struct A} : Type :=
    match A with
      | Func A1 A2 => 
        forall w', w <= w' -> 
                   forall annot', leb annot annot' ->
                                  sforces w' annot' A1 -> Kont sforces w' annot' A2
                                                             
      | Sum A1 A2 => sum (sforces w annot A1) (sforces w annot A2)

      | Nat => X w annot Nat
    end.
  
  (** Monotonicity properties of [sforces] *)
  
  Lemma sforces_mon : forall A w annot, sforces w annot A -> 
                      forall w', w <= w' -> sforces w' annot A.
  Proof.
    induction A.
    
    intros w annot H1 w' ww'.
    simpl in *.
    eauto using wle_trans.
    
    intros w annot H1 w' ww'.
    simpl in *.
    case H1; intro H'.
    eauto using wle_trans.
    eauto using wle_trans.

    simpl.
    intros w a n w' wle'.
    apply X_mon with w.
    assumption.
    assumption.
  Defined.
  
  Lemma sforces_mon2 : forall A w, sforces w false A -> sforces w true A.
  Proof.
    induction A.
    
    intros w H.
    simpl in *.
    intros w' ww' annot' Hannot'.
    apply H.
    assumption.
    constructor.
    
    intros w H1.
    simpl in *.
    destruct H1; auto.

    simpl.
    intros.
    apply X_mon2 with false.
    assumption.
    reflexivity.
  Defined.

  (** The "run" of the continuations monad *)
  Definition run {w annot} :
    Kont sforces w annot Nat -> X w annot Nat.
  Proof.
    intro H.
    destruct annot.
    
    apply H.
    apply wle_refl.
    simpl.
    intros w1 ww1 n.
    exact n.

    apply H.
    apply wle_refl.
    simpl.
    intros w1 ww1 n.
    exact n.
  Defined.
End sforces_cbv.

(** A module type specifying an interpretation of the ground type Nat *)
Module Type Nat_instance_cbv (ks : Kripke_structure).
  Export ks.
  Module ks_monad := Kripke_structure_monad ks.
  Export ks_monad.
  Module cbv_validity := sforces_cbv ks.
  Export cbv_validity.

  Parameter Zero_interpretation : forall w annot, Kont sforces w annot Nat.
  Parameter Succ_interpretation : forall w annot, Kont sforces w annot Nat -> Kont sforces w annot Nat.
  Parameter Rec_interpretation : forall A w annot, Kont sforces w annot Nat -> Kont sforces w annot A -> Kont sforces w annot (Func Nat (Func A A)) -> Kont sforces w annot A.
  Parameter Reset_interpretation : forall w, Kont sforces w true Nat -> Kont sforces w false Nat.
End Nat_instance_cbv.
