type __ = Obj.t
let __ = let rec f _ = Obj.repr f in Obj.repr f

type unit0 =
| Tt

type bool =
| True
| False

type ('a, 'b) sum =
| Inl of 'a
| Inr of 'b

type ('a, 'b) prod =
| Pair of 'a * 'b

(** val fst : ('a1, 'a2) prod -> 'a1 **)

let fst = function
| Pair (x, y) -> x

type 'a list =
| Nil
| Cons of 'a * 'a list

type formula =
| Func of formula * formula
| Sum of formula * formula
| Nat

type context = formula list

type proof =
| Inl0 of context * bool * formula * formula * proof
| Inr0 of context * bool * formula * formula * proof
| Lam of formula list * bool * formula * formula * proof
| Hyp of formula list * bool * formula
| Wkn of context * bool * formula * formula * proof
| Case of context * bool * formula * formula * formula * proof * proof
   * proof
| App of context * bool * formula * formula * proof * proof
| Reset of context * bool * proof
| Shift of formula list * formula * proof
| Zero of context * bool
| Succ of context * bool * proof
| Rec of context * bool * formula * proof * proof * proof

type proof_nf =
| Nf_Inl of context * bool * formula * formula * proof_nf
| Nf_Inr of context * bool * formula * formula * proof_nf
| Nf_Lam of formula list * bool * formula * formula * proof_nf
| Nf_ne of context * bool * formula * proof_ne
| Nf_Zero of context * bool
| Nf_Succ of context * bool * proof_nf
and proof_ne =
| Ne_Hyp of formula list * bool * formula
| Ne_Case of context * bool * formula * formula * formula * proof_ne
   * proof_nf * proof_nf
| Ne_App of context * bool * formula * formula * proof_ne * proof_nf
| Ne_Wkn of context * bool * formula * formula * proof_nf
| Ne_Reset of context * proof_ne
| Ne_Shift of formula list * formula * proof_ne
| Ne_Rec of context * bool * formula * proof_ne * proof_nf * proof_nf

type prefix =
| Prefix_refl of context
| Prefix_cons of context * context * formula * prefix

(** val prefix_trans :
    context -> context -> context -> prefix -> prefix -> prefix **)

let rec prefix_trans w w' w'' h = function
| Prefix_refl w0 -> h
| Prefix_cons (w0, w'0, a, p) ->
  Prefix_cons (w, w'0, a, (prefix_trans w w0 w'0 h p))

(** val proof_ne_mon :
    context -> context -> prefix -> formula -> bool -> proof_ne -> proof_ne **)

let rec proof_ne_mon c c0 p a a0 r =
  match p with
  | Prefix_refl w -> r
  | Prefix_cons (w, w', a1, p0) ->
    Ne_Wkn (w', a0, a, a1, (Nf_ne (w', a0, a,
      (proof_ne_mon w w' p0 a a0 r))))

(** val proof_nf_mon :
    context -> context -> prefix -> formula -> bool -> proof_nf -> proof_nf **)

let rec proof_nf_mon c c0 p a a0 r =
  match p with
  | Prefix_refl w -> r
  | Prefix_cons (w, w', a1, p0) ->
    Nf_ne ((Cons (a1, w')), a0, a, (Ne_Wkn (w', a0, a, a1,
      (proof_nf_mon w w' p0 a a0 r))))

(** val proof_nf_mon2 :
    formula -> context -> bool -> proof_nf -> bool -> proof_nf **)

let rec proof_nf_mon2 a gamma annot p annot' =
  let rec f c b f0 p0 annot'0 =
    match p0 with
    | Nf_Inl (gamma0, annot0, a0, b0, p1) ->
      Nf_Inl (gamma0, annot'0, a0, b0, (f gamma0 annot0 a0 p1 annot'0))
    | Nf_Inr (gamma0, annot0, a0, b0, p1) ->
      Nf_Inr (gamma0, annot'0, a0, b0, (f gamma0 annot0 b0 p1 annot'0))
    | Nf_Lam (gamma0, annot0, a0, b0, p1) ->
      Nf_Lam (gamma0, annot'0, a0, b0,
        (f (Cons (a0, gamma0)) annot0 b0 p1 annot'0))
    | Nf_ne (gamma0, annot0, a0, p1) ->
      Nf_ne (gamma0, annot'0, a0,
        (proof_ne_mon2 a0 gamma0 annot0 p1 annot'0))
    | Nf_Zero (gamma0, annot0) -> Nf_Zero (gamma0, annot'0)
    | Nf_Succ (gamma0, annot0, p1) ->
      Nf_Succ (gamma0, annot'0, (f gamma0 annot0 Nat p1 annot'0))
  in f gamma annot a p annot'

(** val proof_ne_mon2 :
    formula -> context -> bool -> proof_ne -> bool -> proof_ne **)

and proof_ne_mon2 a gamma annot p annot' =
  let rec f c b f0 p0 annot'0 =
    match p0 with
    | Ne_Hyp (gamma0, annot0, a0) -> Ne_Hyp (gamma0, annot'0, a0)
    | Ne_Case (gamma0, annot0, a0, b0, c0, p1, p2, p3) ->
      Ne_Case (gamma0, annot'0, a0, b0, c0,
        (f gamma0 annot0 (Sum (a0, b0)) p1 annot'0),
        (proof_nf_mon2 c0 (Cons (a0, gamma0)) annot0 p2 annot'0),
        (proof_nf_mon2 c0 (Cons (b0, gamma0)) annot0 p3 annot'0))
    | Ne_App (gamma0, annot0, a0, b0, p1, p2) ->
      Ne_App (gamma0, annot'0, a0, b0,
        (f gamma0 annot0 (Func (a0, b0)) p1 annot'0),
        (proof_nf_mon2 a0 gamma0 annot0 p2 annot'0))
    | Ne_Wkn (gamma0, annot0, a0, b0, p1) ->
      Ne_Wkn (gamma0, annot'0, a0, b0,
        (proof_nf_mon2 a0 gamma0 annot0 p1 annot'0))
    | Ne_Reset (gamma0, p1) ->
      (match annot'0 with
       | True -> f gamma0 True Nat p1 True
       | False -> Ne_Reset (gamma0, (f gamma0 True Nat p1 True)))
    | Ne_Shift (gamma0, a0, p1) ->
      Ne_Shift (gamma0, a0,
        (f (Cons ((Func (a0, Nat)), gamma0)) True Nat p1 True))
    | Ne_Rec (gamma0, annot0, c0, p1, p2, p3) ->
      Ne_Rec (gamma0, annot'0, c0, (f gamma0 annot0 Nat p1 annot'0),
        (proof_nf_mon2 c0 gamma0 annot0 p2 annot'0),
        (proof_nf_mon2 (Func (Nat, (Func (c0, c0)))) gamma0 annot0 p3
          annot'0))
  in f gamma annot a p annot'

module type Kripke_structure = 
 sig 
  type coq_K 
  
  type wle 
  
  val wle_refl : coq_K -> wle
  
  val wle_trans : coq_K -> coq_K -> coq_K -> wle -> wle -> wle
  
  type coq_X 
  
  val coq_X_mon : coq_K -> bool -> formula -> coq_X -> coq_K -> wle -> coq_X
  
  val coq_X_mon2 : coq_K -> bool -> formula -> coq_X -> bool -> coq_X
 end

module Kripke_structure_monad = 
 functor (Coq_ks:Kripke_structure) ->
 struct 
  type 'f coq_Kont = __
  
  (** val coq_Kont_mon :
      Coq_ks.coq_K -> bool -> formula -> 'a1 coq_Kont -> Coq_ks.coq_K ->
      Coq_ks.wle -> 'a1 coq_Kont **)
  
  let coq_Kont_mon w annot a h w' ww' =
    match annot with
    | True ->
      Obj.magic (fun w'' w'w'' k ->
        Obj.magic h w'' (Coq_ks.wle_trans w w' w'' ww' w'w'') k)
    | False ->
      Obj.magic (fun c w'' w'w'' k ->
        Obj.magic h c w'' (Coq_ks.wle_trans w w' w'' ww' w'w'') k)
 end

module Coq_forcing_generic_properties = 
 functor (Coq_ks:Kripke_structure) ->
 functor (Coq_sforces_implementation:sig 
  module Coq_ks_monad : 
   sig 
    type 'f coq_Kont = __
    
    val coq_Kont_mon :
      Coq_ks.coq_K -> bool -> formula -> 'a1 coq_Kont -> Coq_ks.coq_K ->
      Coq_ks.wle -> 'a1 coq_Kont
   end
  
  type sforces 
  
  val sforces_mon :
    formula -> Coq_ks.coq_K -> bool -> sforces -> Coq_ks.coq_K -> Coq_ks.wle
    -> sforces
  
  val sforces_mon2 : formula -> Coq_ks.coq_K -> sforces -> sforces
 end) ->
 struct 
  module Coq_ks_monad = Kripke_structure_monad(Coq_ks)
  
  (** val ret :
      Coq_ks.coq_K -> bool -> formula -> Coq_sforces_implementation.sforces
      -> Coq_sforces_implementation.sforces
      Coq_sforces_implementation.Coq_ks_monad.coq_Kont **)
  
  let ret w annot a h =
    match annot with
    | True ->
      Obj.magic (fun w1 wle1 k1 ->
        k1 w1 (Coq_ks.wle_refl w1)
          (Coq_sforces_implementation.sforces_mon a w True h w1 wle1))
    | False ->
      Obj.magic (fun c w1 ww1 k ->
        k w1 (Coq_ks.wle_refl w1)
          (Coq_sforces_implementation.sforces_mon a w False h w1 ww1))
  
  (** val bind :
      Coq_ks.coq_K -> bool -> formula -> formula -> (Coq_ks.coq_K ->
      Coq_ks.wle -> Coq_sforces_implementation.sforces ->
      Coq_sforces_implementation.sforces
      Coq_sforces_implementation.Coq_ks_monad.coq_Kont) ->
      Coq_sforces_implementation.sforces
      Coq_sforces_implementation.Coq_ks_monad.coq_Kont ->
      Coq_sforces_implementation.sforces
      Coq_sforces_implementation.Coq_ks_monad.coq_Kont **)
  
  let bind w annot a b f a0 =
    match annot with
    | True ->
      Obj.magic (fun w1 wle1 k1 ->
        Obj.magic a0 w1 wle1 (fun w2 wle2 a2 ->
          Obj.magic f w2 (Coq_ks.wle_trans w w1 w2 wle1 wle2) a2 w2
            (Coq_ks.wle_refl w2) (fun w3 wle3 b3 ->
            k1 w3 (Coq_ks.wle_trans w1 w2 w3 wle2 wle3) b3)))
    | False ->
      Obj.magic (fun c1 w1 wle1 k1 ->
        Obj.magic a0 c1 w1 wle1 (fun w2 wle2 a2 ->
          Obj.magic f w2 (Coq_ks.wle_trans w w1 w2 wle1 wle2) a2 c1 w2
            (Coq_ks.wle_refl w2) (fun w3 wle3 b3 ->
            k1 w3 (Coq_ks.wle_trans w1 w2 w3 wle2 wle3) b3)))
  
  type sforces_cxt = __
  
  (** val sforces_cxt_mon :
      context -> Coq_ks.coq_K -> bool -> sforces_cxt -> Coq_ks.coq_K ->
      Coq_ks.wle -> sforces_cxt **)
  
  let rec sforces_cxt_mon l w annot h w' h0 =
    match l with
    | Nil -> h
    | Cons (y, l0) ->
      let Pair (s, s0) = Obj.magic h in
      Obj.magic (Pair
        ((Coq_sforces_implementation.sforces_mon y w annot s w' h0),
        (sforces_cxt_mon l0 w annot s0 w' h0)))
  
  (** val sforces_cxt_mon2 :
      context -> Coq_ks.coq_K -> bool -> sforces_cxt -> bool -> sforces_cxt **)
  
  let rec sforces_cxt_mon2 l w annot h annot' =
    match l with
    | Nil -> h
    | Cons (y, l0) ->
      let Pair (s, s0) = Obj.magic h in
      Obj.magic (Pair
        ((match annot' with
          | True ->
            (match annot with
             | True -> s
             | False -> Coq_sforces_implementation.sforces_mon2 y w s)
          | False ->
            (match annot with
             | True -> Obj.magic (fun _ t -> t)
             | False -> s)), (sforces_cxt_mon2 l0 w annot s0 annot')))
 end

module Coq_sforces_cbv = 
 functor (Coq_ks:Kripke_structure) ->
 struct 
  module Coq_ks_monad = Kripke_structure_monad(Coq_ks)
  
  type sforces = __
  
  (** val sforces_mon :
      formula -> Coq_ks.coq_K -> bool -> sforces -> Coq_ks.coq_K ->
      Coq_ks.wle -> sforces **)
  
  let rec sforces_mon f w annot h1 w' ww' =
    match f with
    | Func (f0, f1) ->
      Obj.magic (fun w'0 h annot' _ x0 ->
        Obj.magic h1 w'0 (Coq_ks.wle_trans w w' w'0 ww' h) annot' __ x0)
    | Sum (f0, f1) ->
      (match Obj.magic h1 with
       | Inl h' -> Obj.magic (Inl (sforces_mon f0 w annot h' w' ww'))
       | Inr h' -> Obj.magic (Inr (sforces_mon f1 w annot h' w' ww')))
    | Nat -> Obj.magic (Coq_ks.coq_X_mon w annot Nat (Obj.magic h1) w' ww')
  
  (** val sforces_mon2 : formula -> Coq_ks.coq_K -> sforces -> sforces **)
  
  let rec sforces_mon2 f w h =
    match f with
    | Func (f0, f1) ->
      Obj.magic (fun w' ww' annot' _ -> Obj.magic h w' ww' annot' __)
    | Sum (f0, f1) ->
      (match Obj.magic h with
       | Inl s -> Obj.magic (Inl (sforces_mon2 f0 w s))
       | Inr s -> Obj.magic (Inr (sforces_mon2 f1 w s)))
    | Nat -> Obj.magic (Coq_ks.coq_X_mon2 w False Nat (Obj.magic h) True)
  
  (** val run :
      Coq_ks.coq_K -> bool -> sforces Coq_ks_monad.coq_Kont -> Coq_ks.coq_X **)
  
  let run w annot h =
    match annot with
    | True -> Obj.magic h w (Coq_ks.wle_refl w) (fun w1 ww1 n -> n)
    | False -> Obj.magic h Nat w (Coq_ks.wle_refl w) (fun w1 ww1 n -> n)
 end

module Soundness = 
 functor (Coq_ks:Kripke_structure) ->
 functor (Coq_ni:sig 
  module Coq_ks_monad : 
   sig 
    type 'f coq_Kont = __
    
    val coq_Kont_mon :
      Coq_ks.coq_K -> bool -> formula -> 'a1 coq_Kont -> Coq_ks.coq_K ->
      Coq_ks.wle -> 'a1 coq_Kont
   end
  
  module Coq_cbv_validity : 
   sig 
    module Coq_ks_monad : 
     sig 
      type 'f coq_Kont = __
      
      val coq_Kont_mon :
        Coq_ks.coq_K -> bool -> formula -> 'a1 coq_Kont -> Coq_ks.coq_K ->
        Coq_ks.wle -> 'a1 coq_Kont
     end
    
    type sforces = __
    
    val sforces_mon :
      formula -> Coq_ks.coq_K -> bool -> sforces -> Coq_ks.coq_K ->
      Coq_ks.wle -> sforces
    
    val sforces_mon2 : formula -> Coq_ks.coq_K -> sforces -> sforces
    
    val run :
      Coq_ks.coq_K -> bool -> sforces Coq_ks_monad.coq_Kont -> Coq_ks.coq_X
   end
  
  val coq_Zero_interpretation :
    Coq_ks.coq_K -> bool -> Coq_cbv_validity.sforces
    Coq_cbv_validity.Coq_ks_monad.coq_Kont
  
  val coq_Succ_interpretation :
    Coq_ks.coq_K -> bool -> Coq_cbv_validity.sforces
    Coq_cbv_validity.Coq_ks_monad.coq_Kont -> Coq_cbv_validity.sforces
    Coq_cbv_validity.Coq_ks_monad.coq_Kont
  
  val coq_Rec_interpretation :
    formula -> Coq_ks.coq_K -> bool -> Coq_cbv_validity.sforces
    Coq_cbv_validity.Coq_ks_monad.coq_Kont -> Coq_cbv_validity.sforces
    Coq_cbv_validity.Coq_ks_monad.coq_Kont -> Coq_cbv_validity.sforces
    Coq_cbv_validity.Coq_ks_monad.coq_Kont -> Coq_cbv_validity.sforces
    Coq_cbv_validity.Coq_ks_monad.coq_Kont
  
  val coq_Reset_interpretation :
    Coq_ks.coq_K -> (Coq_ks.coq_K -> Coq_ks.wle -> (Coq_ks.coq_K ->
    Coq_ks.wle -> Coq_ks.coq_X -> Coq_ks.coq_X) -> Coq_ks.coq_X) -> formula
    -> Coq_ks.coq_K -> Coq_ks.wle -> (Coq_ks.coq_K -> Coq_ks.wle ->
    Coq_ks.coq_X -> Coq_ks.coq_X) -> Coq_ks.coq_X
 end) ->
 struct 
  module Coq_ks_monad = Kripke_structure_monad(Coq_ks)
  
  module Coq_generic_properties = Coq_forcing_generic_properties(Coq_ks)(Coq_ni.Coq_cbv_validity)
  
  (** val soundness :
      context -> bool -> formula -> proof -> Coq_ks.coq_K -> bool ->
      Coq_generic_properties.sforces_cxt -> Coq_ni.Coq_cbv_validity.sforces
      Coq_ni.Coq_cbv_validity.Coq_ks_monad.coq_Kont **)
  
  let rec soundness c b f p w annot' wG =
    match p with
    | Inl0 (gamma, annot, a, b0, p0) ->
      (match annot' with
       | True ->
         Obj.magic (fun w1 ww1 k ->
           Obj.magic (fun w0 annot'0 _ wG0 ->
             soundness gamma annot a p0 w0 annot'0 wG0) w True __ wG w1 ww1
             (fun w2 h x0 -> k w2 h (Inl x0)))
       | False ->
         Obj.magic (fun c0 w1 ww1 k ->
           Obj.magic (fun w0 annot'0 _ wG0 ->
             soundness gamma annot a p0 w0 annot'0 wG0) w False __ wG c0 w1
             ww1 (fun w2 h x0 -> k w2 h (Inl x0))))
    | Inr0 (gamma, annot, a, b0, p0) ->
      (match annot' with
       | True ->
         Obj.magic (fun w1 ww1 k ->
           Obj.magic (fun w0 annot'0 _ wG0 ->
             soundness gamma annot b0 p0 w0 annot'0 wG0) w True __ wG w1 ww1
             (fun w2 h x0 -> k w2 h (Inr x0)))
       | False ->
         Obj.magic (fun c0 w1 ww1 k ->
           Obj.magic (fun w0 annot'0 _ wG0 ->
             soundness gamma annot b0 p0 w0 annot'0 wG0) w False __ wG c0 w1
             ww1 (fun w2 h x0 -> k w2 h (Inr x0))))
    | Lam (gamma, annot, a, b0, p0) ->
      Coq_generic_properties.ret w annot' (Func (a, b0))
        (Obj.magic (fun w1 ww1 annot'' _ hA ->
          Obj.magic (fun w0 annot'0 _ wG0 ->
            soundness (Cons (a, gamma)) annot b0 p0 w0 annot'0 wG0) w1
            annot'' __ (Pair (hA,
            (Coq_generic_properties.sforces_cxt_mon gamma w annot''
              (Coq_generic_properties.sforces_cxt_mon2 gamma w annot' wG
                annot'') w1 ww1)))))
    | Hyp (gamma, annot, a) ->
      let Pair (ha, hGamma) = Obj.magic wG in
      Coq_generic_properties.ret w annot' a ha
    | Wkn (gamma, annot, a, b0, p0) ->
      let Pair (wB, wG0) = Obj.magic wG in
      soundness gamma annot a p0 w annot' wG0
    | Case (gamma, annot, a, b0, c0, p0, p1, p2) ->
      (match annot' with
       | True ->
         Obj.magic (fun w1 ww1 k ->
           Obj.magic (fun w0 annot'0 _ wG0 ->
             soundness gamma annot (Sum (a, b0)) p0 w0 annot'0 wG0) w True __
             wG w1 ww1 (fun w2 w1w2 hSum ->
             match hSum with
             | Inl hA ->
               Obj.magic (fun w0 annot'0 _ wG0 ->
                 soundness (Cons (a, gamma)) annot c0 p1 w0 annot'0 wG0) w2
                 True __ (Pair (hA,
                 (Coq_generic_properties.sforces_cxt_mon gamma w True wG w2
                   (Coq_ks.wle_trans w w1 w2 ww1 w1w2)))) w2
                 (Coq_ks.wle_refl w2) (fun w0 h x0 ->
                 k w0 (Coq_ks.wle_trans w1 w2 w0 w1w2 h) x0)
             | Inr hB ->
               Obj.magic (fun w0 annot'0 _ wG0 ->
                 soundness (Cons (b0, gamma)) annot c0 p2 w0 annot'0 wG0) w2
                 True __ (Pair (hB,
                 (Coq_generic_properties.sforces_cxt_mon gamma w True wG w2
                   (Coq_ks.wle_trans w w1 w2 ww1 w1w2)))) w2
                 (Coq_ks.wle_refl w2) (fun w0 h x0 ->
                 k w0 (Coq_ks.wle_trans w1 w2 w0 w1w2 h) x0)))
       | False ->
         Obj.magic (fun d w1 ww1 k ->
           Obj.magic (fun w0 annot'0 _ wG0 ->
             soundness gamma annot (Sum (a, b0)) p0 w0 annot'0 wG0) w False
             __ wG d w1 ww1 (fun w2 w1w2 hSum ->
             match hSum with
             | Inl hA ->
               Obj.magic (fun w0 annot'0 _ wG0 ->
                 soundness (Cons (a, gamma)) annot c0 p1 w0 annot'0 wG0) w2
                 False __ (Pair (hA,
                 (Coq_generic_properties.sforces_cxt_mon gamma w False wG w2
                   (Coq_ks.wle_trans w w1 w2 ww1 w1w2)))) d w2
                 (Coq_ks.wle_refl w2) (fun w0 h x0 ->
                 k w0 (Coq_ks.wle_trans w1 w2 w0 w1w2 h) x0)
             | Inr hB ->
               Obj.magic (fun w0 annot'0 _ wG0 ->
                 soundness (Cons (b0, gamma)) annot c0 p2 w0 annot'0 wG0) w2
                 False __ (Pair (hB,
                 (Coq_generic_properties.sforces_cxt_mon gamma w False wG w2
                   (Coq_ks.wle_trans w w1 w2 ww1 w1w2)))) d w2
                 (Coq_ks.wle_refl w2) (fun w0 h x0 ->
                 k w0 (Coq_ks.wle_trans w1 w2 w0 w1w2 h) x0))))
    | App (gamma, annot, a, b0, p0, p1) ->
      (match annot' with
       | True ->
         Obj.magic (fun w1 ww1 k ->
           Obj.magic (fun w0 annot'0 _ wG0 ->
             soundness gamma annot (Func (a, b0)) p0 w0 annot'0 wG0) w1 True
             __
             (Coq_generic_properties.sforces_cxt_mon gamma w True wG w1 ww1)
             w1 (Coq_ks.wle_refl w1) (fun w2 w1w2 hAB ->
             Obj.magic (fun w0 annot'0 _ wG0 ->
               soundness gamma annot a p1 w0 annot'0 wG0) w True __ wG w2
               (Coq_ks.wle_trans w w1 w2 ww1 w1w2) (fun w3 w2w3 hA ->
               hAB w3 w2w3 True __ hA w3 (Coq_ks.wle_refl w3)
                 (fun w4 w3w4 hB' ->
                 k w4
                   (Coq_ks.wle_trans w1 w2 w4 w1w2
                     (Coq_ks.wle_trans w2 w3 w4 w2w3 w3w4)) hB'))))
       | False ->
         Obj.magic (fun c0 w1 ww1 k ->
           Obj.magic (fun w0 annot'0 _ wG0 ->
             soundness gamma annot (Func (a, b0)) p0 w0 annot'0 wG0) w1 False
             __
             (Coq_generic_properties.sforces_cxt_mon gamma w False wG w1 ww1)
             c0 w1 (Coq_ks.wle_refl w1) (fun w2 w1w2 hAB ->
             Obj.magic (fun w0 annot'0 _ wG0 ->
               soundness gamma annot a p1 w0 annot'0 wG0) w False __ wG c0 w2
               (Coq_ks.wle_trans w w1 w2 ww1 w1w2) (fun w3 w2w3 hA ->
               hAB w3 w2w3 False __ hA c0 w3 (Coq_ks.wle_refl w3)
                 (fun w4 w3w4 hB' ->
                 k w4
                   (Coq_ks.wle_trans w1 w2 w4 w1w2
                     (Coq_ks.wle_trans w2 w3 w4 w2w3 w3w4)) hB')))))
    | Reset (gamma, annot, p0) ->
      (match annot' with
       | True ->
         let iHp' = soundness gamma True Nat p0 w True wG in
         let iHp'0 = Coq_ni.Coq_cbv_validity.run w True iHp' in
         Coq_generic_properties.ret w True Nat (Obj.magic iHp'0)
       | False ->
         Obj.magic
           (Coq_ni.coq_Reset_interpretation w
             (let wGamma' =
                Coq_generic_properties.sforces_cxt_mon2 gamma w False wG True
              in
              let iHp' = soundness gamma True Nat p0 w True wGamma' in
              Obj.magic iHp')))
    | Shift (gamma, a, p0) ->
      let x0 = fun w1 ww1 kK ->
        Obj.magic (fun w0 annot'0 _ wG0 ->
          soundness (Cons ((Func (a, Nat)), gamma)) True Nat p0 w0 annot'0
            wG0) w1 True __ (Pair ((fun w2 w1w2 annot'' _ hA w3 w2w3 kK' ->
          kK' w3 (Coq_ks.wle_refl w3)
            (kK w3 (Coq_ks.wle_trans w1 w2 w3 w1w2 w2w3)
              (Coq_ni.Coq_cbv_validity.sforces_mon a w2 True hA w3 w2w3))),
          (Coq_generic_properties.sforces_cxt_mon gamma w True wG w1 ww1)))
          w1 (Coq_ks.wle_refl w1) (fun w2 w1w2 n -> n)
      in
      (fun w1 -> Obj.magic x0 w1)
    | Zero (gamma, annot) -> Coq_ni.coq_Zero_interpretation w annot'
    | Succ (gamma, annot, p0) ->
      Coq_ni.coq_Succ_interpretation w annot'
        (soundness gamma annot Nat p0 w annot' wG)
    | Rec (gamma, annot, c0, p0, p1, p2) ->
      Coq_ni.coq_Rec_interpretation c0 w annot'
        (soundness gamma annot Nat p0 w annot' wG)
        (soundness gamma annot c0 p1 w annot' wG)
        (soundness gamma annot (Func (Nat, (Func (c0, c0)))) p2 w annot' wG)
 end

module Completeness = 
 struct 
  module U = 
   struct 
    type coq_K = context
    
    type wle = prefix
    
    (** val wle_refl : context -> prefix **)
    
    let wle_refl x =
      Prefix_refl x
    
    (** val wle_trans :
        context -> context -> context -> prefix -> prefix -> prefix **)
    
    let wle_trans =
      prefix_trans
    
    type coq_X = proof_nf
    
    (** val coq_X_mon :
        context -> bool -> formula -> coq_X -> context -> wle -> coq_X **)
    
    let coq_X_mon w annot a h w' ww' =
      proof_nf_mon w w' ww' a annot h
    
    (** val coq_X_mon2 :
        context -> bool -> formula -> coq_X -> bool -> coq_X **)
    
    let coq_X_mon2 w annot a h annot' =
      proof_nf_mon2 a w annot h annot'
   end
  
  module Coq_ks_monad = Kripke_structure_monad(U)
  
  module Coq_cbv_validity = Coq_sforces_cbv(U)
  
  module Coq_generic_properties = Coq_forcing_generic_properties(U)(Coq_cbv_validity)
  
  (** val completeness :
      formula -> context -> bool -> (Coq_cbv_validity.sforces
      Coq_cbv_validity.Coq_ks_monad.coq_Kont -> proof_nf, proof_ne ->
      Coq_cbv_validity.sforces Coq_cbv_validity.Coq_ks_monad.coq_Kont) prod **)
  
  let rec completeness = function
  | Func (f0, f1) ->
    let iHA1 = completeness f0 in
    let iHA2 = completeness f1 in
    (fun w annot -> Pair ((fun h -> Nf_Lam (w, annot, f0, f1,
    (let h0 = fun w0 annot0 -> let Pair (x, x0) = iHA2 w0 annot0 in x in
     Obj.magic h0 (Cons (f0, w)) annot (fun w1 incl1 k1 ->
       match annot with
       | True ->
         let h1 = let Pair (x, x0) = iHA1 w1 True in x0 in
         Obj.magic h1
           (proof_ne_mon (Cons (f0, w)) w1 incl1 f0 True (Ne_Hyp (w, True,
             f0))) w1 (U.wle_refl w1) (fun w2 incl2 h2 ->
           Obj.magic h w2
             (U.wle_trans w w1 w2
               (U.wle_trans w (Cons (f0, w)) w1 (Prefix_cons (w, w, f0,
                 (Prefix_refl w))) incl1) incl2) (fun w3 incl3 f3 ->
             f3 w3 (U.wle_refl w3) True __
               (Coq_cbv_validity.sforces_mon f0 w2 True h2 w3 incl3) w3
               (U.wle_refl w3) (fun w4 incl4 h4 ->
               k1 w4
                 (U.wle_trans w1 w3 w4 (U.wle_trans w1 w2 w3 incl2 incl3)
                   incl4) h4)))
       | False ->
         (fun k2 ->
           let Pair (x, x0) = Obj.magic iHA1 incl1 False in
           x0
             (proof_ne_mon (Cons (f0, w)) (Obj.magic incl1) (Obj.magic k1) f0
               False (Ne_Hyp (w, False, f0))) w1 incl1 (Prefix_refl
             (Obj.magic incl1)) (fun w2 incl2 h2 ->
             Obj.magic h w1 w2
               (U.wle_trans w (Obj.magic incl1) w2
                 (U.wle_trans w (Cons (f0, w)) (Obj.magic incl1) (Prefix_cons
                   (w, w, f0, (Prefix_refl w))) (Obj.magic k1)) incl2)
               (fun w3 incl3 f3 ->
               f3 w3 (U.wle_refl w3) False __
                 (Coq_cbv_validity.sforces_mon f0 w2 False h2 w3 incl3) w1 w3
                 (U.wle_refl w3) (fun w4 incl4 h4 ->
                 k2 w4
                   (U.wle_trans (Obj.magic incl1) w3 w4
                     (U.wle_trans (Obj.magic incl1) w2 w3 incl2 incl3) incl4)
                   h4)))))))), (fun e ->
    Coq_generic_properties.ret w annot (Func (f0, f1))
      (Obj.magic (fun w' incl' annot' _ h' ->
        let h'0 = Coq_generic_properties.ret w' annot' f0 h' in
        let h = let Pair (x, x0) = iHA1 w' annot' in x in
        let h'1 = h h'0 in
        let Pair (x, x0) = iHA2 w' annot' in
        x0 (Ne_App (w', annot', f0, f1,
          (proof_ne_mon w w' incl' (Func (f0, f1)) annot'
            (proof_ne_mon2 (Func (f0, f1)) w annot e annot')), h'1)))))))
  | Sum (f0, f1) ->
    let iHA1 = completeness f0 in
    let iHA2 = completeness f1 in
    (fun w annot -> Pair ((fun h ->
    match annot with
    | True ->
      Nf_ne (w, True, (Sum (f0, f1)), (Ne_Shift (w, (Sum (f0, f1)), (Ne_App
        ((Cons ((Func ((Sum (f0, f1)), Nat)), w)), True, (Sum (f0, f1)), Nat,
        (Ne_Hyp (w, True, (Func ((Sum (f0, f1)), Nat)))), (Nf_Inl ((Cons
        ((Func ((Sum (f0, f1)), Nat)), w)), True, f0, f1,
        (let h0 = fun w0 annot0 -> let Pair (x, x0) = iHA1 w0 annot0 in x in
         Obj.magic h0 (Cons ((Func ((Sum (f0, f1)), Nat)), w)) True
           (fun w1 incl1 h1 -> Nf_ne (w1, True, Nat, (Ne_App (w1, True, (Sum
           (f0, f1)), Nat,
           (proof_ne_mon (Cons ((Func ((Sum (f0, f1)), Nat)), w)) w1 incl1
             (Func ((Sum (f0, f1)), Nat)) True (Ne_Hyp (w, True, (Func ((Sum
             (f0, f1)), Nat))))), (Nf_Inr (w1, True, f0, f1,
           (let h2 = fun w0 annot0 -> let Pair (x, x0) = iHA2 w0 annot0 in x
            in
            Obj.magic h2 w1 True (fun w2 incl2 h3 ->
              Obj.magic h w2
                (U.wle_trans w w1 w2
                  (U.wle_trans w (Cons ((Func ((Sum (f0, f1)), Nat)), w)) w1
                    (Prefix_cons (w, w, (Func ((Sum (f0, f1)), Nat)),
                    (U.wle_refl w))) incl1) incl2) (fun w3 incl3 h4 ->
                match h4 with
                | Inl s -> h1 w3 (U.wle_trans w1 w2 w3 incl2 incl3) s
                | Inr s -> h3 w3 incl3 s)))))))))))))))))
    | False ->
      Obj.magic h (Sum (f0, f1)) w (U.wle_refl w) (fun w1 incl1 h1 ->
        match h1 with
        | Inl h11 ->
          Nf_Inl (w1, False, f0, f1,
            (let annot0 = False in
             let Pair (x, x0) = iHA1 w1 annot0 in
             x (Coq_generic_properties.ret w1 False f0 h11)))
        | Inr h12 ->
          Nf_Inr (w1, False, f0, f1,
            (let annot0 = False in
             let Pair (x, x0) = iHA2 w1 annot0 in
             x (Coq_generic_properties.ret w1 False f1 h12))))), (fun e ->
    match annot with
    | True ->
      Obj.magic (fun w1 incl1 k1 -> Nf_ne (w1, True, Nat, (Ne_Case (w1, True,
        f0, f1, Nat, (proof_ne_mon w w1 incl1 (Sum (f0, f1)) True e),
        (let h = let Pair (x, x0) = iHA1 (Cons (f0, w1)) True in x0 in
         Obj.magic h (Ne_Hyp (w1, True, f0)) (Cons (f0, w1))
           (U.wle_refl (Cons (f0, w1))) (fun w2 incl2 h2 ->
           k1 w2
             (U.wle_trans w1 (Cons (f0, w1)) w2 (Prefix_cons (w1, w1, f0,
               (Prefix_refl w1))) incl2) (Inl h2))),
        (let h = let Pair (x, x0) = iHA2 (Cons (f1, w1)) True in x0 in
         Obj.magic h (Ne_Hyp (w1, True, f1)) (Cons (f1, w1))
           (U.wle_refl (Cons (f1, w1))) (fun w2 incl2 h2 ->
           k1 w2
             (U.wle_trans w1 (Cons (f1, w1)) w2 (Prefix_cons (w1, w1, f1,
               (Prefix_refl w1))) incl2) (Inr h2)))))))
    | False ->
      Obj.magic (fun c1 w1 incl1 k1 -> Nf_ne (w1, False, c1, (Ne_Case (w1,
        False, f0, f1, c1, (proof_ne_mon w w1 incl1 (Sum (f0, f1)) False e),
        (let h = let Pair (x, x0) = iHA1 (Cons (f0, w1)) False in x0 in
         Obj.magic h (Ne_Hyp (w1, False, f0)) c1 (Cons (f0, w1))
           (U.wle_refl (Cons (f0, w1))) (fun w2 incl2 h2 ->
           k1 w2
             (U.wle_trans w1 (Cons (f0, w1)) w2 (Prefix_cons (w1, w1, f0,
               (Prefix_refl w1))) incl2) (Inl h2))),
        (let h = let Pair (x, x0) = iHA2 (Cons (f1, w1)) False in x0 in
         Obj.magic h (Ne_Hyp (w1, False, f1)) c1 (Cons (f1, w1))
           (U.wle_refl (Cons (f1, w1))) (fun w2 incl2 h2 ->
           k1 w2
             (U.wle_trans w1 (Cons (f1, w1)) w2 (Prefix_cons (w1, w1, f1,
               (Prefix_refl w1))) incl2) (Inr h2))))))))))
  | Nat ->
    (fun w annot -> Pair ((fun h -> Coq_cbv_validity.run w annot h),
      (fun e ->
      Coq_generic_properties.ret w annot Nat
        (Obj.magic (Nf_ne (w, annot, Nat, e))))))
  
  module NI = 
   struct 
    module Coq_ks_monad = Kripke_structure_monad(U)
    
    module Coq_cbv_validity = Coq_sforces_cbv(U)
    
    (** val coq_Zero_interpretation :
        context -> bool -> Coq_cbv_validity.sforces
        Coq_cbv_validity.Coq_ks_monad.coq_Kont **)
    
    let coq_Zero_interpretation w annot =
      Coq_generic_properties.ret w annot Nat (Obj.magic (Nf_Zero (w, annot)))
    
    (** val coq_Succ_interpretation :
        context -> bool -> Coq_cbv_validity.sforces
        Coq_cbv_validity.Coq_ks_monad.coq_Kont -> Coq_cbv_validity.sforces
        Coq_cbv_validity.Coq_ks_monad.coq_Kont **)
    
    let coq_Succ_interpretation w annot r =
      let r0 = Coq_cbv_validity.run w annot r in
      Coq_generic_properties.ret w annot Nat
        (Obj.magic (Nf_Succ (w, annot, r0)))
    
    (** val reset_nf : context -> proof_nf -> proof_nf **)
    
    let rec reset_nf gamma = function
    | Nf_ne (gamma0, annot, a, h) ->
      Nf_ne (gamma, False, Nat, (Ne_Reset (gamma, h)))
    | Nf_Zero (gamma0, annot) -> Nf_Zero (gamma, False)
    | Nf_Succ (gamma0, annot, h) ->
      Nf_Succ (gamma, False, (reset_nf gamma h))
    | _ -> assert false (* absurd case *)
    
    (** val coq_Reset_interpretation :
        context -> (context -> U.wle -> (context -> U.wle -> U.coq_X ->
        U.coq_X) -> U.coq_X) -> formula -> context -> U.wle -> (context ->
        U.wle -> U.coq_X -> U.coq_X) -> U.coq_X **)
    
    let coq_Reset_interpretation w h =
      let h0 = Coq_cbv_validity.run w True (Obj.magic h) in
      Obj.magic
        (Coq_generic_properties.ret w False Nat
          (let h1 = fun _ _ _ ->
             match h0 with
             | Nf_ne (gamma, annot, a, h1) ->
               Nf_ne (w, False, Nat, (Ne_Reset (w, h1)))
             | Nf_Zero (gamma, annot) -> Nf_Zero (w, False)
             | Nf_Succ (gamma, annot, h1) -> reset_nf w h1
             | _ -> assert false (* absurd case *)
           in
           Obj.magic h1 __ __ __))
    
    (** val coq_Rec_interpretation :
        formula -> context -> bool -> Coq_cbv_validity.sforces
        Coq_cbv_validity.Coq_ks_monad.coq_Kont -> Coq_cbv_validity.sforces
        Coq_cbv_validity.Coq_ks_monad.coq_Kont -> Coq_cbv_validity.sforces
        Coq_cbv_validity.Coq_ks_monad.coq_Kont -> Coq_cbv_validity.sforces
        Coq_cbv_validity.Coq_ks_monad.coq_Kont **)
    
    let coq_Rec_interpretation a w annot n a0 f =
      let n0 = Coq_cbv_validity.run w annot n in
      (match n0 with
       | Nf_ne (gamma, annot0, a1, p) ->
         (match annot0 with
          | True ->
            Obj.magic (fun w1 leq1 k1 -> Nf_ne (w1, True, Nat, (Ne_Rec (w1,
              True, Nat, (proof_ne_mon gamma w1 leq1 Nat True p),
              (Obj.magic a0 w1 leq1 k1),
              (let h =
                 Coq_generic_properties.bind (Cons (Nat, (Cons (Nat, w1))))
                   True (Func (Nat, (Func (a, a)))) a (fun gamma' leq' phi ->
                   Coq_generic_properties.bind gamma' True (Func (a, a)) a
                     (fun w' h x0 ->
                     Coq_generic_properties.bind w' True a a
                       (fun w'0 h0 x1 ->
                       Coq_generic_properties.ret w'0 True a x1)
                       (Coq_cbv_validity.Coq_ks_monad.coq_Kont_mon gamma True
                         a a0 w'
                         (U.wle_trans gamma gamma' w'
                           (U.wle_trans gamma (Cons (Nat, (Cons (Nat, w1))))
                             gamma'
                             (U.wle_trans gamma w1 (Cons (Nat, (Cons (Nat,
                               w1)))) leq1 (Prefix_cons (w1, (Cons (Nat,
                               w1)), Nat, (Prefix_cons (w1, w1, Nat,
                               (Prefix_refl w1)))))) leq') h)))
                     (Obj.magic phi gamma' (U.wle_refl gamma') True __
                       (U.coq_X_mon gamma True Nat (Nf_ne (gamma, True, Nat,
                         p)) gamma'
                         (U.wle_trans gamma (Cons (Nat, (Cons (Nat, w1))))
                           gamma'
                           (U.wle_trans gamma w1 (Cons (Nat, (Cons (Nat,
                             w1)))) leq1 (Prefix_cons (w1, (Cons (Nat, w1)),
                             Nat, (Prefix_cons (w1, w1, Nat, (Prefix_refl
                             w1)))))) leq'))))
                   (Coq_cbv_validity.Coq_ks_monad.coq_Kont_mon gamma True
                     (Func (Nat, (Func (a, a)))) f (Cons (Nat, (Cons (Nat,
                     w1))))
                     (U.wle_trans gamma w1 (Cons (Nat, (Cons (Nat, w1))))
                       leq1 (Prefix_cons (w1, (Cons (Nat, w1)), Nat,
                       (Prefix_cons (w1, w1, Nat, (Prefix_refl w1)))))))
               in
               Nf_Lam (w1, True, Nat, (Func (Nat, Nat)), (Nf_Lam ((Cons (Nat,
               w1)), True, Nat, Nat,
               (Obj.magic h (Cons (Nat, (Cons (Nat, w1)))) (Prefix_refl (Cons
                 (Nat, (Cons (Nat, w1))))) (fun w2 leq2 h2 ->
                 k1 w2
                   (U.wle_trans w1 (Cons (Nat, (Cons (Nat, w1)))) w2
                     (Prefix_cons (w1, (Cons (Nat, w1)), Nat, (Prefix_cons
                     (w1, w1, Nat, (Prefix_refl w1))))) leq2) h2))))))))))
          | False ->
            Obj.magic (fun c1 w1 leq1 k1 -> Nf_ne (w1, False, c1, (Ne_Rec
              (w1, False, c1, (proof_ne_mon gamma w1 leq1 Nat False p),
              (Obj.magic a0 c1 w1 leq1 k1),
              (let h =
                 Coq_generic_properties.bind (Cons (c1, (Cons (Nat, w1))))
                   False (Func (Nat, (Func (a, a)))) a
                   (fun gamma' leq' phi ->
                   Coq_generic_properties.bind gamma' False (Func (a, a)) a
                     (fun w' h x0 ->
                     Coq_generic_properties.bind w' False a a
                       (fun w'0 h0 x1 ->
                       Coq_generic_properties.ret w'0 False a x1)
                       (Coq_cbv_validity.Coq_ks_monad.coq_Kont_mon gamma
                         False a a0 w'
                         (U.wle_trans gamma gamma' w'
                           (U.wle_trans gamma (Cons (c1, (Cons (Nat, w1))))
                             gamma'
                             (U.wle_trans gamma w1 (Cons (c1, (Cons (Nat,
                               w1)))) leq1 (Prefix_cons (w1, (Cons (Nat,
                               w1)), c1, (Prefix_cons (w1, w1, Nat,
                               (Prefix_refl w1)))))) leq') h)))
                     (Obj.magic phi gamma' (U.wle_refl gamma') False __
                       (U.coq_X_mon gamma False Nat (Nf_ne (gamma, False,
                         Nat, p)) gamma'
                         (U.wle_trans gamma (Cons (c1, (Cons (Nat, w1))))
                           gamma'
                           (U.wle_trans gamma w1 (Cons (c1, (Cons (Nat,
                             w1)))) leq1 (Prefix_cons (w1, (Cons (Nat, w1)),
                             c1, (Prefix_cons (w1, w1, Nat, (Prefix_refl
                             w1)))))) leq'))))
                   (Coq_cbv_validity.Coq_ks_monad.coq_Kont_mon gamma False
                     (Func (Nat, (Func (a, a)))) f (Cons (c1, (Cons (Nat,
                     w1))))
                     (U.wle_trans gamma w1 (Cons (c1, (Cons (Nat, w1)))) leq1
                       (Prefix_cons (w1, (Cons (Nat, w1)), c1, (Prefix_cons
                       (w1, w1, Nat, (Prefix_refl w1)))))))
               in
               Nf_Lam (w1, False, Nat, (Func (c1, c1)), (Nf_Lam ((Cons (Nat,
               w1)), False, c1, c1,
               (Obj.magic h c1 (Cons (c1, (Cons (Nat, w1)))) (Prefix_refl
                 (Cons (c1, (Cons (Nat, w1))))) (fun w2 leq2 h2 ->
                 k1 w2
                   (U.wle_trans w1 (Cons (c1, (Cons (Nat, w1)))) w2
                     (Prefix_cons (w1, (Cons (Nat, w1)), c1, (Prefix_cons
                     (w1, w1, Nat, (Prefix_refl w1))))) leq2) h2)))))))))))
       | Nf_Zero (gamma, annot0) -> a0
       | Nf_Succ (gamma, annot0, n1) ->
         Coq_generic_properties.bind gamma annot0 (Func (Nat, (Func (a, a))))
           a (fun gamma' leq' phi ->
           Coq_generic_properties.bind gamma' annot0 (Func (a, a)) a
             (fun w' h x0 ->
             Coq_generic_properties.bind w' annot0 a a (fun w'0 h0 x1 ->
               Obj.magic x0 w'0 h0 annot0 __ x1)
               (Coq_cbv_validity.Coq_ks_monad.coq_Kont_mon gamma annot0 a a0
                 w' (U.wle_trans gamma gamma' w' leq' h)))
             (Obj.magic phi gamma' (U.wle_refl gamma') annot0 __
               (U.coq_X_mon gamma annot0 Nat n1 gamma' leq'))) f
       | _ -> Obj.magic (fun _ t -> t))
   end
  
  module Coq_soundness_for_U = Soundness(U)(NI)
  
  (** val coq_NbE : bool -> formula -> proof -> proof_nf **)
  
  let coq_NbE annot a p =
    let compl = fst (completeness a Nil annot) in
    let hnil = Tt in
    let sndns =
      Coq_soundness_for_U.soundness Nil annot a p Nil annot (Obj.magic hnil)
    in
    compl sndns
 end

