(** * Call-by-VALUE TDPE for shift and reset and higher type recursion over natural numbers (System T) *)

Require Import syntax.
Require Import model_structure.
Require Import model_validity_generic.
Require Import model_validity_cbv.

Set Implicit Arguments.
Unset Strict Implicit.

(** ** Evaluator of typed terms in Kripke models (aka Soundness) *)
Module Soundness (ks : Kripke_structure)(ni : Nat_instance_cbv ks).
  Export ks.
  Module ks_monad := Kripke_structure_monad ks.
  Export ks_monad.
  Export ni.
  Module generic_properties := forcing_generic_properties ks cbv_validity.
  Export generic_properties.

  (* soundness can be proved more nicely if I reuse the proof I wrote
     for nf_to_Kont_Nat below *)
  Theorem soundness : 
    forall Gamma annot A, proof Gamma annot A -> 
      forall w, forall annot', leb annot annot' -> 
        w ||++ Gamma ; annot' -> w ||- A ; annot'.
  Proof.
    intros Gamma annot A p.
    induction p.

    intros w annot' Hleb wG.
    destruct annot'.
      intros w1 ww1 k.
      unfold Kont in IHp.
      apply (IHp w true).
      assumption.
      assumption.
      assumption.
      intros.
      apply k.
      assumption.
      simpl.
      left;assumption.
      intros C w1 ww1 k.
      unfold Kont in IHp.
      apply (IHp w false).
      assumption.
      assumption.
      assumption.
      intros.
      apply k.
      assumption.
      simpl.
      left;assumption.

    intros w annot' Hleb wG.
    destruct annot'.
      intros w1 ww1 k.
      unfold Kont in IHp.
      apply (IHp w true).
      assumption.
      assumption.
      assumption.
      intros.
      apply k.
      assumption.
      simpl.
      right;assumption.
      intros C w1 ww1 k.
      unfold Kont in IHp.
      apply (IHp w false).
      assumption.
      assumption.
      assumption.
      intros.
      apply k.
      assumption.
      simpl.
      right;assumption.

    intros w annot' Hleb wG.
    apply ret.
    simpl.
    intros w1 ww1 annot'' Hleb'' HA.
    simpl in IHp.
    apply IHp.
    eauto using leb_trans.
    split;
      eauto using sforces_cxt_mon,sforces_cxt_mon2.

    simpl.
    intros w annot' Hleb [Ha HGamma].
    apply ret.
    assumption.

    intros w annot' Hleb [wB wG].
    apply IHp.
    assumption.
    assumption.
    
    clear p1 p2 p3.
    intros w annot' Hleb' wGamma.
    destruct annot'.
      intros w1 ww1 k.
      unfold Kont in IHp1.
      apply (IHp1 w true); clear IHp1.
      assumption.
      assumption.
      assumption.
      intros w2 w1w2 HSum.
      simpl in HSum.
      case HSum.
      intro HA.
      unfold Kont in IHp2.
      apply (IHp2 w2 true).
      assumption.
      simpl.
      split.
      assumption.
      eauto using wle_trans,sforces_cxt_mon.
      apply wle_refl.
      intros; eauto using wle_trans.
      intro HB.
      unfold Kont in IHp3.
      apply (IHp3 w2 true).
      assumption.
      simpl.
      split.
      assumption.
      eauto using wle_trans,sforces_cxt_mon.
      apply wle_refl.
      intros; eauto using wle_trans.
      intros D w1 ww1 k.
      unfold Kont in IHp1.
      apply (IHp1 w false); clear IHp1.
      assumption.
      assumption.
      assumption.
      intros w2 w1w2 HSum.
      simpl in HSum.
      case HSum.
      intro HA.
      unfold Kont in IHp2.
      apply (IHp2 w2 false).
      assumption.
      simpl.
      split.
      assumption.
      eauto using wle_trans,sforces_cxt_mon.
      apply wle_refl.
      intros; eauto using wle_trans.
      intro HB.
      unfold Kont in IHp3.
      apply (IHp3 w2 false).
      assumption.
      simpl.
      split.
      assumption.
      eauto using wle_trans,sforces_cxt_mon.
      apply wle_refl.
      intros; eauto using wle_trans.

    intros w annot' Hleb' wGamma.
    destruct annot'.
      intros w1 ww1 k.
      unfold Kont in IHp1.
      apply (IHp1 w1 true).
      assumption.
      eauto using sforces_cxt_mon.
      apply wle_refl.
      simpl sforces.
      intros w2 w1w2 HAB.
      unfold Kont in IHp2.
      apply (IHp2 w true).
      assumption.
      assumption.
      eauto using wle_trans.
      intros w3 w2w3 HA.
      unfold Kont in HAB.
      apply (HAB w3 w2w3 true).
      reflexivity.
      assumption.
      apply wle_refl.
      intros w4 w3w4 HB'.
      apply k.
      eauto using wle_trans.
      assumption.
      intros C w1 ww1 k.
      unfold Kont in IHp1.
      apply (IHp1 w1 false).
      assumption.
      eauto using sforces_cxt_mon.
      apply wle_refl.
      simpl sforces.
      intros w2 w1w2 HAB.
      unfold Kont in IHp2.
      apply (IHp2 w false).
      assumption.
      assumption.
      eauto using wle_trans.
      intros w3 w2w3 HA.
      unfold Kont in HAB.
      apply (HAB w3 w2w3 false).
      constructor.
      assumption.
      apply wle_refl.
      intros w4 w3w4 HB'.
      apply k.
      eauto using wle_trans.
      assumption.

    intros w annot' Hleb' wGamma.
    destruct annot'.
    assert (IHp' := IHp w true (leb_refl _) wGamma).
    apply run in IHp'.
    apply ret.
    simpl.
    apply IHp'.
    apply Reset_interpretation.
    assert (wGamma' : w ||++ Gamma; true).
    apply sforces_cxt_mon2 with false.
    assumption.
    reflexivity.
    assert (IHp' := IHp w true (leb_refl _) wGamma').
    apply IHp'.

    intros w annot' Hleb' wGamma.
    inversion Hleb' as [Heq].
    rewrite Heq in *.
    intros w1 ww1 KK.
    (* apply run. *)
    apply (IHp w1 true).
    reflexivity.
    split.
    simpl.
    intros w2 w1w2 annot'' Heq'' HA.
    rewrite Heq'' in *.
    intros w3 w2w3 KK'.
    apply KK'.
    apply wle_refl.
    simpl.
    apply KK.
    eauto using wle_trans.
    eauto using sforces_mon.
    eauto using sforces_cxt_mon.
    apply wle_refl.
    simpl.
    intros w2 w1w2 n.
    exact n.

    intros.
    apply Zero_interpretation.

    intros.
    apply Succ_interpretation.
    apply IHp; assumption.

    intros.
    apply Rec_interpretation.
    apply IHp1; assumption.
    apply IHp2; assumption.
    apply IHp3; assumption.
  Defined.
End Soundness.

(** ** The reifyer from Kripke models to terms in normal form (aka Completeness for the Universal model) *)
Module Completeness.

  (** The "Universal" Kripke model [U] build from syntax *)
  Module U <: Kripke_structure.

    Definition K := context.

    Definition wle := prefix.

    Notation "w <= w'" := (wle w w').

    Definition wle_refl := prefix_refl.

    Definition wle_trans := prefix_trans.

    Definition X (w:context)(annot:bool)(A:formula) : Set :=  
      proof_nf w annot A.
    
    Lemma X_mon : forall w annot A, X w annot A -> 
                  forall w', w <= w' -> X w' annot A.
    Proof.
      intros w annot A H w' ww'.
      apply proof_nf_mon with w.
      assumption.
      assumption.
    Defined.
    
    Lemma X_mon2 : forall w annot A, X w annot A -> 
                   forall annot', leb annot annot' -> X w annot' A.
    Proof.
      intros w annot A H annot' Hleb.
      apply proof_nf_mon2 with annot.
      assumption.
      assumption.
    Defined.
  End U.

  Export U.
  Module ks_monad := Kripke_structure_monad U.
  Export ks_monad.
  Module cbv_validity := sforces_cbv U.
  Export cbv_validity.
  Module generic_properties := forcing_generic_properties U cbv_validity.
  Export generic_properties.

  Theorem completeness : forall A w annot,
      (Kont sforces w annot A -> proof_nf w annot A) *
      (proof_ne w annot A -> Kont sforces w annot A).
  Proof.
    induction A.

    (* ARROW TYPE *)
    intros.
    split.
    (* REIFY *)
    intro H.
    apply nf_Lam.
    apply IHA2.
    unfold Kont.
    destruct annot.
    (* classical case *)
    intros w1 incl1 k1.
    apply (IHA1 w1 true).
    apply proof_ne_mon with (A1::w).
    assumption.
    apply ne_Hyp.
    apply wle_refl.
    intros w2 incl2 H2.
    apply H.
    apply wle_trans with w1.
    apply wle_trans with (A1 :: w).
    constructor.
    constructor.
    assumption.
    assumption.
    intros w3 incl3 f3.
    simpl in f3.
    apply (f3 w3 (wle_refl w3) true).
    reflexivity.
    apply sforces_mon with w2; auto.
    apply wle_refl.
    intros w4 incl4 H4.
    apply k1.
    apply wle_trans with w3;
      eauto using wle_trans.
    assumption.
    (* intuitionistic case *)
    intros C w1 incl1 k1.
    apply (IHA1 w1 false).
    apply proof_ne_mon with (A1::w).
    assumption.
    apply ne_Hyp.
    constructor.
    intros w2 incl2 H2.
    apply H.
    apply wle_trans with w1.
    apply wle_trans with (A1 :: w).
    constructor.
    constructor.
    assumption.
    assumption.
    intros w3 incl3 f3.
    simpl in f3.
    apply (f3 w3 (wle_refl w3) false).
    constructor.
    apply sforces_mon with w2; auto.
    apply wle_refl.
    intros w4 incl4 H4.
    apply k1.
    apply wle_trans with w3;
      eauto using wle_trans.
    assumption.
    (* REFLECT *)
    intro e.
    apply ret.
    simpl.
    intros w' incl' annot' leb' H'.
    apply ret in H'.
    apply (IHA1 w' annot') in H'.
    apply (IHA2 w' annot').
    apply ne_App with A1.
    apply proof_ne_mon with (Gamma:=w).
    assumption.
    apply proof_ne_mon2 with annot.
    assumption.
    assumption.
    assumption.

    (* SUM TYPE *)
    intros.
    split.
    (* REIFY *)
    intro H.
    destruct annot.
    (* classical case *)
    apply nf_ne.
    apply ne_Shift.
    apply ne_App with (Sum A1 A2).
    apply ne_Hyp.
    apply nf_Inl.
    apply IHA1.
    intros w1 incl1 H1.
    apply nf_ne.
    apply ne_App with (Sum A1 A2).
    eapply proof_ne_mon.
    apply incl1.
    apply ne_Hyp.
    apply nf_Inr.
    apply IHA2.
    intros w2 incl2 H2.
    apply H.
    assert (w <= Func (Sum A1 A2) Nat :: w).
    right.
    apply wle_refl.
    apply wle_trans with w1.
    apply wle_trans with (Func (Sum A1 A2) Nat :: w).
    right.
    apply wle_refl.
    assumption.
    assumption.
    intros w3 incl3 H3.
    destruct H3.
    apply H1.
    apply wle_trans with w2; assumption.
    assumption.
    apply H2.
    assumption.
    assumption.
    (* intuitionistic case *)
    unfold Kont in H.
    apply (H (Sum A1 A2)).
    apply wle_refl.
    simpl.
    intros w1 incl1 H1.
    destruct H1 as [H11 | H12].
    apply nf_Inl.
    apply IHA1.
    apply ret.
    assumption.
    apply nf_Inr.
    apply IHA2.
    apply ret.
    assumption.
    (* REFLECT *)
    intro e.
    destruct annot.
    (* classical case *)
    intros w1 incl1 k1.
    apply nf_ne.
    apply ne_Case with A1 A2.
    eauto using proof_ne_mon.
    apply (IHA1 (A1::w1) true).
    apply ne_Hyp.
    apply wle_refl.
    intros w2 incl2 H2.
    apply k1.
    apply wle_trans with (A1::w1).
    constructor.
    constructor.
    assumption.
    simpl.
    left.
    assumption.
    apply (IHA2 (A2::w1) true).
    apply ne_Hyp.
    apply wle_refl.
    intros w2 incl2 H2.
    apply k1.
    apply wle_trans with (A2::w1).
    constructor.
    constructor.
    assumption.
    simpl.
    right.
    assumption.
    (* intuitionistic case *)
    intros C1 w1 incl1 k1.
    apply nf_ne.
    apply ne_Case with A1 A2.
    eauto using proof_ne_mon.
    apply (IHA1 (A1::w1) false).
    apply ne_Hyp.
    apply wle_refl.
    intros w2 incl2 H2.
    apply k1.
    apply wle_trans with (A1::w1).
    constructor.
    constructor.
    assumption.
    simpl.
    left.
    assumption.
    apply (IHA2 (A2::w1) false).
    apply ne_Hyp.
    apply wle_refl.
    intros w2 incl2 H2.
    apply k1.
    apply wle_trans with (A2::w1).
    constructor.
    constructor.
    assumption.
    simpl.
    right.
    assumption.

    (* NAT TYPE *)
    intros.
    split.
    (* REIFY *)
    intro H.
    apply run in H.
    unfold X in H.
    apply H.
    (* REFLECT *)
    intro e.
    apply ret.
    apply nf_ne.
    exact e.
  Defined.

  Module NI <: Nat_instance_cbv U.
    Module ks_monad := Kripke_structure_monad U.
    Export ks_monad.
    Module cbv_validity := sforces_cbv U.
    Export cbv_validity.

    Lemma Zero_interpretation : forall w annot, Kont sforces w annot Nat.
    Proof.
      intros.
      apply ret.
      apply nf_Zero.
    Defined.

    Lemma Succ_interpretation : forall w annot, Kont sforces w annot Nat -> Kont sforces w annot Nat.
    Proof.
      intros w annot r.
      apply run in r.
      apply ret.
      apply nf_Succ.
      apply r.
    Defined.

    Fixpoint reset_nf (Gamma:context)(r:proof_nf Gamma true Nat) {struct r} : proof_nf Gamma false Nat.
    Proof.
      inversion r.
      Guarded.
      { apply nf_ne.
        apply ne_Reset.
        assumption. }
      { apply nf_Zero. }
      { apply nf_Succ.
        apply reset_nf.
        assumption. }
      Guarded.
    Defined.

    Lemma Reset_interpretation : forall w, Kont sforces w true Nat -> Kont sforces w false Nat.
    Proof.
      intros w H.
      apply run in H.
      apply ret.
      simpl.
      unfold X in *.
      inversion H.
      apply nf_ne.
      apply ne_Reset.
      assumption.
      apply nf_Zero.
      apply reset_nf.
      apply H0.
    Defined.

    Lemma Rec_interpretation : forall A w annot, Kont sforces w annot Nat -> Kont sforces w annot A -> Kont sforces w annot (Func Nat (Func A A)) -> Kont sforces w annot A.
    Proof.
      intros A w annot n a f.
      apply run in n.
      unfold X in n.
      (* inversion n. *)
      remember Nat as fla.
      (* induction n; *)
      destruct n;
        try (clear -Heqfla ; congruence);
        try rewrite Heqfla in *;
        clear Heqfla.
      (* nf_ne *)
      {destruct annot.
       - intros w1 leq1 k1.
         apply nf_ne.
         apply ne_Rec.
         + apply proof_ne_mon with Gamma.
           assumption.
           apply p.
         + apply a.
           assumption.
           apply k1.
         + assert (H : (Nat :: Nat :: w1) ||- A; true).
           { apply (bind (A:=Func Nat (Func A A))).
             intros Gamma' leq' phi.
             apply (bind (A:=Func A A)).
             intros.
             apply (bind (A:=A)).
             intros.
             apply ret.
             assumption.
             apply Kont_mon with Gamma.
             apply a.
             apply wle_trans with Gamma'.
             apply wle_trans with (Nat::Nat::w1).
             apply wle_trans with w1.
             assumption.
             right;right;apply prefix_refl.
             assumption.
             assumption.
             apply phi.
             apply wle_refl.
             apply leb_refl.
             apply X_mon with Gamma.
             apply nf_ne.
             apply p.
             apply wle_trans with (Nat::Nat::w1).
             apply wle_trans with w1.
             assumption.
             right;right;apply prefix_refl.
             assumption.
             apply Kont_mon with Gamma.
             apply f.
             apply wle_trans with w1.
             assumption.
             right;right;apply prefix_refl. }
           unfold Kont in H.
           unfold X in H.
           apply nf_Lam.
           apply nf_Lam.
           apply H.
           apply prefix_refl.
           intros w2 leq2 H2.
           apply k1.
           apply wle_trans with (Nat::Nat::w1).
           right;right;apply prefix_refl.
           assumption.        
           assumption.        
       - intros C1 w1 leq1 k1.
         apply nf_ne.
         apply ne_Rec.
         + apply proof_ne_mon with Gamma.
           assumption.
           apply p.
         + apply a.
           assumption.
           apply k1.
         + assert (H : (C1 :: Nat :: w1) ||- A; false).
           { apply (bind (A:=Func Nat (Func A A))).
             intros Gamma' leq' phi.
             apply (bind (A:=Func A A)).
             intros.
             apply (bind (A:=A)).
             intros.
             apply ret.
             assumption.
             apply Kont_mon with Gamma.
             apply a.
             apply wle_trans with Gamma'.
             apply wle_trans with (C1::Nat::w1).
             apply wle_trans with w1.
             assumption.
             right;right;apply prefix_refl.
             assumption.
             assumption.
             apply phi.
             apply wle_refl.
             apply leb_refl.
             apply X_mon with Gamma.
             apply nf_ne.
             apply p.
             apply wle_trans with (C1::Nat::w1).
             apply wle_trans with w1.
             assumption.
             right;right;apply prefix_refl.
             assumption.
             apply Kont_mon with Gamma.
             apply f.
             apply wle_trans with w1.
             assumption.
             right;right;apply prefix_refl. }
           unfold Kont in H.
           unfold X in H.
           apply nf_Lam.
           apply nf_Lam.
           apply H.
           apply prefix_refl.
           intros w2 leq2 H2.
           apply k1.
           apply wle_trans with (C1::Nat::w1).
           right;right;apply prefix_refl.
           assumption.        
           assumption.        
      }
      (* nf_Zero *)
      apply a.
      (* nf_Succ *)
      apply (bind (A:=Func Nat (Func A A))).
      intros Gamma' leq' phi.
      apply (bind (A:=Func A A)).
      intros.
      apply (bind (A:=A)).
      intros.
      apply X0.
      assumption.
      apply leb_refl.
      assumption.
      apply Kont_mon with Gamma.
      apply a.
      eapply wle_trans.
      apply leq'.
      assumption.
      apply phi.
      apply wle_refl.
      apply leb_refl.
      apply X_mon with Gamma.
      apply n.
      assumption.
      apply f.
    Defined.
  End NI.

  (** ** Now we can define NbE as composition of soundness and completeness *)
  Module soundness_for_U := Soundness U NI.

  Check soundness_for_U.soundness.

  Definition NbE (annot:bool)(A:formula)(p:proof nil annot A) : proof_nf nil annot A.
  Proof.
    (* begin show *)
    set (compl := fst (completeness A nil annot)).
    set (empty_cxt := nil).
    assert (Hnil : sforces_cxt empty_cxt annot empty_cxt).
    simpl.
    constructor.
    set (sndns := soundness_for_U.soundness p (w:=nil) (leb_refl _) Hnil).
    exact (compl sndns).
  (* end show *)
  Defined.


End Completeness.

Extraction "nbe_cbv" Completeness.NbE.
